#!/usr/bin/python

"""

An very partial implementation of an EPP client in Python.

Implements:

* most of RFC 4930
* a small part of RFC 4931 (Domain Mapping)
* RFC 4934 (TCP transport, other transports are not supported)

Stephane Bortzmeyer <bortzmeyer@nic.fr>

"""

import socket
import sys
import struct
import random
import time
try:
    import cElementTree as ElementTree
except ImportError:
    try:
        import ElementTree
    except ImportError:
        try:
            import xml.etree.ElementTree as ElementTree
        except ImportError:
            fatal("You need the ElementTree module, available at " + \
                  "http://effbot.org/zone/element-index.htm or in Python >= 2.5")

class EPP_Exception(Exception):
    pass

# Class methods
def make_epp(xml):
    epp = ElementTree.Element("epp", xmlns=EPP.NS)
    epp.append(xml)
    return epp

def epp_element(name, ns=None):
    if ns is None:
        return ElementTree.Element("%s" % name)
    else:
        return ElementTree.Element("%s" % name, xmlns=ns)
            
def check(assertion, message):
    if not assertion:
        raise EPP_Exception("EPP protocol error \"%s\"" % message)

def int_from_net(data):
    return struct.unpack(format_32, data)[0]

def int_to_net(value):
    return struct.pack(format_32, value)

def safe_recv(s, n):
    data = ''
    while (n > 0):
        tmp = s.recv(n)
        data += tmp
        n -= len(tmp)
    return data
   
class EPP:

    NS = "urn:ietf:params:xml:ns:epp-1.0"

    def __init__(self, server, port, me, mypassword):
        self._server = server
        self._port = port
        self._id = me
        self._password = mypassword
        self.debug = 0
        self._generator = random.Random (time.time())
        AF_UNSPEC = 0
        addrs = socket.getaddrinfo(self._server, self._port,
                                   AF_UNSPEC, socket.SOCK_STREAM)
        connected = False
        for addr in addrs:
            address_family, socktype, proto, canonname, sa = addr
            self._socket = socket.socket(address_family, socktype, proto)
            try:
                self._socket.connect(sa)
                connected = True
            except socket.error, e:
                error = "Failure to connect to %s: %s" % (sa[0], e)
                continue
            break
        if not connected:
            raise EPP_Exception("Cannot connect to %s:%i (%s)" % \
                                (self._server, self._port, error))
        if self.debug >= 2:
            print "Connected to %s:%i (%s)" % (self._server, self._port, sa[0])
        self.handle_greeting()
       
    def random_transaction(self):
        number = self._generator.randrange (3, 64)
        # Min and max values for trIDStringType in the EPP schema
        name = ""
        for i in range (number):
            name = name + chr (self._generator.randrange (65, 90))
            # TODO: digits are accepted, too
        return name

    def read_epp(self):
        data = self._socket.recv(4) # RFC 4934, section 4
        length = int_from_net(data) 
        if self.debug >= 3:
            print "XML Element of %i bytes" % length
        data = safe_recv(self._socket, length-4)
        epp = ElementTree.fromstring(data)
        if epp.tag != "{%s}epp" % EPP.NS:
            raise EPP_Exception("Not an EPP instance: %s" % epp.tag)
        xml = epp[0]
        if self.debug == 3:
            print "XML Element <%s>" % xml.tag
        if self.debug > 3:
            print "Received:\n%s" % ElementTree.tostring(xml)
        return xml

    def send_epp(self, xml):
        # TODO: we should separate transport issues (such as the
        # 4-bytes length prepended) from XML issues, in order to
        # support other protocols).
        epp = make_epp(xml)
        epp_as_string = ElementTree.tostring(epp, encoding="UTF-8")
        length = int_to_net(len(epp_as_string) + 4 + 2)
        if self.debug == 3:
            print "Sending \"%s\"" % epp[0]
        if self.debug > 3:
            print "Sending \"%s\"" % epp_as_string
        self._socket.send(length)
        self._socket.send(epp_as_string + "\r\n")

    def make_command(self, mycommand):
        command = epp_element("command")
        command.append(mycommand)
        transactionID = epp_element("clTRID")
        transaction = self.random_transaction()
        transactionID.text = transaction
        command.append(transactionID)
        return (command, transaction)
    
    def handle_greeting(self):
        hello = epp_element("hello")
        self.send_epp(hello)
        xml = self.read_epp()
        check(xml.tag == "{%s}greeting" % EPP.NS,
              "No greeting, %s" % xml.tag)

    def log_in(self):
        login = epp_element("login")
        clID = epp_element("clID")
        clID.text = self._id
        password = epp_element("pw")
        password.text = self._password
        login.append(clID)
        login.append(password)
        options = epp_element("options")
        version = epp_element("version")
        version.text = "1.0"
        options.append(version)
        lang = epp_element("lang")
        lang.text = "fr"
        options.append(lang)
        login.append(options)
        domain = epp_element("objURI")
        domain.text = EPP.Domain.NS
        svcs = epp_element("svcs")
        svcs.append(domain)
        login.append(svcs)
        (command, trID) = self.make_command(login)
        self.send_epp(command)
        xml = self.read_epp()
        check(xml.tag == "{%s}response" % EPP.NS,
              "No response but %s" % xml.tag)
        result = xml[0]
        check(result.tag == "{%s}result" % EPP.NS,
              "No result but %s" % result.tag)
        code = result.attrib['code']
        if code != "1000":
            raise EPP_Exception("Cannot log in, result is %s" % code)

    def log_out(self):
        logout = epp_element("logout")
        (command, trID) = self.make_command(logout)
        self.send_epp(command)
        xml = self.read_epp()
        check(xml.tag == "{%s}response" % EPP.NS,
              "No response but %s" % xml.tag)
        result = xml[0]
        check(result.tag == "{%s}result" % EPP.NS,
              "No result but %s" % result.tag)
        code = result.attrib['code']
        if code != "1500":
            raise EPP_Exception("Cannot log out properly, result is %s" % code)

    class Domain:
        NS = "urn:ietf:params:xml:ns:domain-1.0"

        def __init__(self, server, name):
            self.name = name
            self._server = server
    
        def check(self):
            name = epp_element("name")
            name.text = self.name
            dcheck = epp_element("check", ns=EPP.Domain.NS)
            dcheck.append(name)
            ccheck = epp_element("check", ns=EPP.NS)
            ccheck.append(dcheck)
            (command, trID) = self._server.make_command(ccheck)
            self._server.send_epp(command)
            xml = self._server.read_epp()
            check(xml.tag == "{%s}response" % EPP.NS,
                  "No response but %s" % xml.tag)
            result = xml[0]
            check(result.tag == "{%s}result" % EPP.NS,
                  "No result but %s" % result.tag)
            code = result.attrib['code']
            if code != "1000":
                raise EPP_Exception("Cannot check, result is %s" % code)
            resdata = xml[1]
            domaindata = resdata[0]
            domaincd = domaindata[0]
            check(domaincd.tag == "{%s}cd" % EPP.Domain.NS,
                  "No cd but %s" % domaincd.tag)
            domainname = domaincd[0]
            if len(domaincd) > 1:
                reason = domaincd[1]
            check(domainname.text == self.name,
                  "Wrong domain name, received: %s, expected: %s" % \
                  (domainname.text, self.name))
            available = domainname.attrib["avail"]
            if available == "1":
                return (True, "")
            else:
                return (False, reason.text)
            

# Initialize some stuff

format_32 = ">I" # Forcing the endianness with > also forces the size. Yes.
